"""This module has been moved to Aries.storage package.
The imports in this files are for backward compatibility purpose.
Use Aries.storage module for new developments.
This file will be removed in the future.
"""
from ..storage.gs import GSObject, GSFolder, GSFile
